#ifndef _ORTHONORMALIZE_KERNEL_
#define _ORTHONORMALIZE_KERNEL_
/*

height should be equal to blockDim.y

*/
#include <math.h>
/*
The following function implements the stabilized Gram–Schmidt orthonormalization.
Matrix "a" will contain the orthogonal matrix "Q" at the end.
*/
__device__ unsigned int flag = 0; // number of orthogonalized columns
__global__ void orthonormalize(float* a) {
	__shared__ float s[BLOCK_DIM];
	int tid = threadIdx.x;
	int idx = blockIdx.x * blockDim.x + tid;
	float me;
	int i = 0;
	while (i < blockIdx.x) {
		if (flag > i) {
			float vi = a[i * blockDim.x + tid];
	 		me = a[idx];
	 		s[tid] = vi * me;
	 		__syncthreads();
	 		int k = blockDim.x / 2;
 			while (k != 0) {
				if(tid < k)
					s[tid] += s[tid + k];
				__syncthreads();
				k /= 2;
			}
			if (blockDim.x % 2 != 0 && tid == 0)
				s[0] += s[blockDim.x - 1];
	 		__syncthreads();
	 		a[idx] -= s[0] * vi;
	 		__threadfence_block();
	 		i++;
		}
	}
	me = a[idx];
	s[tid] = me * me;
	__syncthreads();
	i = blockDim.x / 2;
	while (i != 0) {
		if(tid < i) {
			s[tid] += s[tid + i];
		}
		__syncthreads();
		i /= 2;
	}
	if (blockDim.x % 2 != 0 && tid == 0)
	{
		s[0] += s[blockDim.x - 1];
	}
	__syncthreads();
	a[idx] = me / sqrtf(s[0]);
	__threadfence();
	if (tid == 0) flag++;
	__threadfence();
}

#endif
